#define VERSION Version: 0.99, last modified 15 January 94: ALPHA
#define PROGNAME "vmstat"
/* Copyright 1994 by Henry Ware <al172@yfn.ysu.edu>. Copyleft same year. */
/* Copyright 2002 Rik van Riel <riel@conectiva.com.br */
/* This attempts to display virtual memory statistics.
   vmstat does not need to be run as root.
*/
/* TODO etc.
 * Count the system calls.  How?  I don't even know where to start.
 * add more items- aim at Posix 1003.2 (1003.7?) compliance, if relevant (I.  
   don't think it is, but let me know.)
 * sometimes d(inter)<d(ticks)!!  This is not possible: inter is a sum of 
   positive numbers including ticks.  But it happens.  This doesn't seem to
   affect things much, however.
 * It would be interesting to see when the buffers avert a block io:
   Like SysV4's "sar -b"... it might fit better here, with Linux's variable 
   buffer size.
 * Ideally, blocks in & out would be counted in 1k increments, rather than
   by block: this only makes a difference for CDs and is a problematic fix.
*/
/* PROCPS
   The procps suite is maintained by Michael K. Johnson <johnsonm@redhat.com>
*/

#include "proc/sysinfo.h"
#include "proc/version.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/dir.h>
#include <dirent.h>

#define NDEBUG !DEBUG

#define BUFFSIZE 8192
#define FALSE 0
#define TRUE 1

int main(int, char **);
void usage(void);
void crash(char *);
int winhi(void);
void showheader(void);
void getstat(unsigned *, unsigned *, unsigned *, unsigned *,
	     unsigned *, unsigned *, unsigned *, unsigned *,
	     unsigned *, unsigned *, unsigned *, unsigned *);
void getmeminfo(unsigned *, unsigned *, unsigned *, unsigned *, unsigned int);
void getrunners(unsigned *, unsigned *);
static char buff[BUFFSIZE];	/* used in the procedures */
static unsigned int actinact = FALSE;

/***************************************************************
                             Main 
***************************************************************/

int
main(int argc, char *argv[])
{

	const char format[] =
	    "%2u %2u %6u %6u %6u %6u %4u %4u %5u %5u %4u %5u %2u %2u %2u %2u\n";
	unsigned int height = 22;	/* window height, reset later if needed. */
	unsigned int moreheaders = TRUE;
	unsigned int tog = 0;	/* toggle switch for cleaner code */
	unsigned int i, hz;
	unsigned int running, blocked;
	unsigned int cpu_use[2], cpu_nic[2], cpu_sys[2];
	unsigned int duse, dsys, didl, diow, divi, divo2;
	unsigned int memfree, membuff, swapused, memcache;
	unsigned int cpu_idl[2];
	unsigned int cpu_iow[2];
	unsigned int pgpgin[2], pgpgout[2], pswpin[2], pswpout[2];
	unsigned int inter[2], ticks[2], ctxt[2];
	unsigned int per = 0, pero2;
	unsigned long num = 0;
	unsigned int kb_per_page = sysconf(_SC_PAGESIZE) / 1024;

	setlinebuf(stdout);
	argc = 0;		/* redefined as number of integer arguments */
	per = 1;
	num = 0;
	for (argv++; *argv; argv++) {
		if ('-' == (**argv)) {
			switch (*(++(*argv))) {
			case 'V':
				display_version();
				exit(0);
			case 'n':
				/* print only one header */
				moreheaders = FALSE;
			case 'a':
				/* Print active/inactive instead of buff/cache */
				actinact = TRUE;
				break;
			default:
				/* no other aguments defined yet. */
				usage();
			}
		} else {
			argc++;
			switch (argc) {
			case 1:
				if ((per = atoi(*argv)) == 0)
					usage();
				num = ULONG_MAX;
				break;
			case 2:
				num = atol(*argv);
				break;
			default:
				usage();
			}	/* switch */
		}
	}

	if (moreheaders) {
		int tmp = winhi() - 3;
		height = ((tmp > 0) ? tmp : 22);
	}

	pero2 = (per / 2);
	showheader();
	getrunners(&running, &blocked);
	getmeminfo(&memfree, &membuff, &swapused, &memcache, actinact);
	getstat(cpu_use, cpu_nic, cpu_sys, cpu_idl, cpu_iow,
		pgpgin, pgpgout, pswpin, pswpout, inter, ticks, ctxt);
	duse = *(cpu_use) + *(cpu_nic);
	dsys = *(cpu_sys);
	didl = *(cpu_idl);
	diow = *(cpu_iow);
	divi = (duse + dsys + didl + diow);
	if (!divi)
		divi = 1;
	hz = sysconf(_SC_CLK_TCK);	/* get ticks/s from system */
	divo2 = divi / 2;
	printf(format,
	       running, blocked,
	       swapused, memfree, membuff, memcache,
	       (*(pswpin) * kb_per_page * hz + divo2) / divi,
	       (*(pswpout) * kb_per_page * hz + divo2) / divi,
	       (*(pgpgin) * hz + divo2) / divi,
	       (*(pgpgout) * hz + divo2) / divi,
	       (*(inter) * hz + divo2) / divi,
	       (*(ctxt) * hz + divo2) / divi,
	       (100 * duse + divo2) / divi, (100 * dsys + divo2) / divi,
	       (100 * didl + divo2) / divi,
	       (100 * diow + divo2) / divi);

	for (i = 1; i < num; i++) {	/* \\\\\\\\\\\\\\\\\\\\ main loop ////////////////// */
		sleep(per);
		if (moreheaders && ((i % height) == 0))
			showheader();
		tog = !tog;
		getrunners(&running, &blocked);
		getmeminfo(&memfree, &membuff, &swapused, &memcache, actinact);
		getstat(cpu_use + tog, cpu_nic + tog, cpu_sys + tog,
			cpu_idl + tog, cpu_iow + tog, pgpgin + tog, pgpgout + tog,
			pswpin + tog, pswpout + tog, inter + tog, ticks + tog,
			ctxt + tog);
		duse =
		    *(cpu_use + tog) - *(cpu_use + !tog) + *(cpu_nic + tog) -
		    *(cpu_nic + !tog);
		dsys = *(cpu_sys + tog) - *(cpu_sys + !tog);
		didl = *(cpu_idl+!tog) <= *(cpu_idl+tog) ? /* test overflow */
			(*(cpu_idl + tog) - *(cpu_idl + !tog)) % UINT_MAX : 0;
		diow = *(cpu_iow + tog) - *(cpu_iow + !tog);
		divi = (duse + dsys + didl + diow);
		if (!divi)
			divi = 1;
		divo2 = divi / 2;
		printf(format,
		       running, blocked,
		       swapused, memfree, membuff, memcache,
		       (((*(pswpin + tog) - *(pswpin + (!tog))) * kb_per_page +
			 pero2) / per),
		       ((*(pswpout + tog) - *(pswpout + (!tog))) * kb_per_page +
			pero2) / per,
		       (*(pgpgin + tog) - *(pgpgin + (!tog)) + pero2) / per,
		       (*(pgpgout + tog) - *(pgpgout + (!tog)) + pero2) / per,
		       (*(inter + tog) - *(inter + (!tog)) + pero2) / per,
		       (*(ctxt + tog) - *(ctxt + (!tog)) + pero2) / per,
		       (100 * duse + divo2) / divi, (100 * dsys + divo2) / divi,
		       (100 * didl + divo2) / divi,
		       (100 * diow + divo2) / divi);
	}
	exit(EXIT_SUCCESS);
}

/**************************** others ***********************************/

void
usage(void)
{
	fprintf(stderr, "usage: %s [-V] [-n] [-a] [delay [count]]\n", PROGNAME);
	fprintf(stderr, "              -V prints version.\n");
	fprintf(stderr, "              -n causes the headers not to be reprinted regularly.\n");
	fprintf(stderr, "              -a print active/inactive page stats.\n");
	fprintf(stderr,
		"              delay is the delay between updates in seconds. \n");
	fprintf(stderr, "              count is the number of updates.\n");
	exit(EXIT_FAILURE);
}

void
crash(char *filename)
{
	perror(filename);
	exit(EXIT_FAILURE);
}

int
winhi(void)
{
	struct winsize win;
	int rows = 24;

	if (ioctl(1, TIOCGWINSZ, &win) != -1 && win.ws_row > 0)
		rows = win.ws_row;

	return rows;
}

void
showheader(void)
{
	printf("%5s%28s%10s%12s%11s%12s\n",
	       "procs", "memory", "swap", "io", "system", "cpu");
	if (actinact)
		printf
		    ("%2s %2s %6s %6s %6s %6s %4s %4s %5s %5s %4s %5s %2s %2s %2s %2s\n",
		     "r", "b", "swpd", "free", "inact", "active", "si",
		     "so", "bi", "bo", "in", "cs", "us", "sy", "id", "wa");
	else
		printf
		    ("%2s %2s %6s %6s %6s %6s %4s %4s %5s %5s %4s %5s %2s %2s %2s %2s\n",
		     "r", "b", "swpd", "free", "buff", "cache", "si", "so",
		     "bi", "bo", "in", "cs", "us", "sy", "id", "wa");
}

static void
procstat(unsigned *cuse, unsigned *cice, unsigned *csys,
	 unsigned *cide, unsigned *ciow, unsigned *pin, unsigned *pout,
	 unsigned *swin, unsigned *swout, unsigned *itot, unsigned *i1,
	 unsigned *ct)
{
	static int statfd;

	if ((statfd = open("/proc/stat", O_RDONLY, 0)) != -1) {
		char *b;
		unsigned cirq=0, csoftirq=0;
		buff[BUFFSIZE - 1] = 0;	/* ensure null termination in buffer */
		read(statfd, buff, BUFFSIZE - 1);
		close(statfd);
		*itot = 0;
		*i1 = 1;	/* ensure assert below will fail if the sscanf bombs */
		b = strstr(buff, "cpu ");
		if (b) {
			sscanf(b, "cpu  %u %u %u %u %u %u %u", cuse, cice, csys, cide, ciow, &cirq, &csoftirq);
			*csys += cirq + csoftirq;
		}
		b = strstr(buff, "page ");
		if (b)
			sscanf(b, "page %u %u", pin, pout);
		b = strstr(buff, "swap ");
		if (b)
			sscanf(b, "swap %u %u", swin, swout);
		b = strstr(buff, "intr ");
		if (b)
			sscanf(b, "intr %u %u", itot, i1);
		b = strstr(buff, "ctxt ");
		if (b)
			sscanf(b, "ctxt %u", ct);
	} else {
		crash("/proc/stat");
	}
}

/* For 2.5 kernels */
static void
scan_one(char *key, unsigned *val)
{
	char *b = strstr(buff, key);
	if (b) {
		b = strstr(b, " ");
		if (b)
			sscanf(b, " %u", val);
	}
}

static void
procvmstat(unsigned *pin, unsigned *pout, unsigned *swin, unsigned *swout)
{
	int fd;

	fd = open("/proc/vmstat", O_RDONLY, 0);
	if (fd < 0)
		return;
	read(fd, buff, BUFFSIZE);
	buff[BUFFSIZE - 1] = 0;
	close(fd);
	scan_one("pgpgin", pin);
	scan_one("pgpgout", pout);
	scan_one("pswpin", swin);
	scan_one("pswpout", swout);
}

void
getstat(unsigned *cuse, unsigned *cice, unsigned *csys,
	unsigned *cide, unsigned *ciow, unsigned *pin, unsigned *pout, unsigned *swin,
	unsigned *swout, unsigned *itot, unsigned *i1, unsigned *ct)
{
	*cuse = 0;
	*cice = 0;
	*csys = 0;
	*cide = 0;
	*ciow = 0;
	*pin = 0;
	*pout = 0;
	*swin = 0;
	*swout = 0;
	*itot = 0;
	*i1 = 0;
	*ct = 0;
	procstat(cuse, cice, csys, cide, ciow, pin, pout, swin, swout, itot, i1, ct);
	procvmstat(pin, pout, swin, swout);
}

void
getmeminfo(unsigned *memfree, unsigned *membuff, unsigned *swapused,
	   unsigned *memcache, unsigned int activeinactive)
{
	struct meminfo_struct mem_info = MEMINFO_ZERO;
	meminfo(&mem_info);
	if (mem_info.mem.total == 0)
		crash("couldn't parse /proc/meminfo");
	*memfree = mem_info.mem.free >> 10;	/* bytes to kB */
	*swapused = mem_info.swap.used >> 10;
	if (activeinactive) {
		/* Get alternate values instead... */
		*membuff = mem_info.mem.inactive_dirty >> 10;
		*membuff += mem_info.mem.inactive_clean >> 10;
		*memcache = mem_info.mem.active >> 10;
	} else {
		*membuff = mem_info.mem.buffers >> 10;
		*memcache = mem_info.mem.cached >> 10;
	}
}

void
getrunners(unsigned int *running, unsigned int *blocked)
{
	static struct direct *ent;
	static char filename[80];
	static int fd;
	static unsigned size;
	static char c;
	DIR *proc;

	*running = 0;
	*blocked = 0;

	if ((proc = opendir("/proc")) == NULL)
		crash("/proc");

	while ((ent = readdir(proc))) {
		if (isdigit(ent->d_name[0])) {	/*just to weed out the obvious junk */
			sprintf(filename, "/proc/%s/stat", ent->d_name);
			if ((fd = open(filename, O_RDONLY, 0)) != -1) {	/*this weeds the rest */
				read(fd, buff, BUFFSIZE - 1);
				sscanf(buff,
				       "%*d %*s %c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %*d %*d %*d %*d %*d %*d %*u %*u %*d %*u %u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u\n",
				       &c, &size);
				close(fd);

				if (c == 'R') {
					(*running)++;
				} else if (c == 'D') {
					(*blocked)++;
				}
			}
		}
	}
	closedir(proc);

#if 1
	/* is this next line a good idea?  It removes this thing which
	   uses (hopefully) little time, from the count of running processes */
	(*running)--;
#endif
}
